@@ -1,5 +1,5 @@
 /*
- * Copyright 2009-2012 10gen, Inc.
+ * Copyright 2009-2014 MongoDB, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -540,12 +540,50 @@ static PyObject* _cbson_get_more_message(PyObject* self, PyObject* args) {
     return result;
 }
 
+static void
+_set_document_too_large(int size, long max) {
+    PyObject* DocumentTooLarge = _error("DocumentTooLarge");
+    if (DocumentTooLarge) {
+#if PY_MAJOR_VERSION >= 3
+        PyObject* error = PyUnicode_FromFormat(DOC_TOO_LARGE_FMT, size, max);
+#else
+        PyObject* error = PyString_FromFormat(DOC_TOO_LARGE_FMT, size, max);
+#endif
+        if (error) {
+            PyErr_SetObject(DocumentTooLarge, error);
+            Py_DECREF(error);
+        }
+        Py_DECREF(DocumentTooLarge);
+    }
+}
+
+static PyObject*
+_send_insert(PyObject* self, PyObject* client,
+             PyObject* gle_args, buffer_t buffer,
+             char* coll_name, int coll_len, int request_id, int safe) {
+
+    PyObject* result;
+    if (safe) {
+        if (!add_last_error(self, buffer, request_id,
+                            coll_name, coll_len, gle_args)) {
+            return NULL;
+        }
+    }
+
+    result = Py_BuildValue("i" BYTES_FORMAT_STRING, request_id,
+                           buffer_get_buffer(buffer),
+                           buffer_get_position(buffer));
+
+    return PyObject_CallMethod(client, "_send_message", "NN",
+                               result, PyBool_FromLong((long)safe));
+}
+
 static PyObject* _cbson_do_batched_insert(PyObject* self, PyObject* args) {
     struct module_state *state = GETSTATE(self);
 
     /* NOTE just using a random number as the request_id */
     int request_id = rand();
-    int options = 0, max_size = 0;
+    int send_safe, options = 0;
     int length_location, message_length;
     int collection_name_length;
     char* collection_name = NULL;
@@ -557,11 +595,11 @@ static PyObject* _cbson_do_batched_insert(PyObject* self, PyObject* args) {
     PyObject* result;
     PyObject* max_bson_size_obj;
     PyObject* max_message_size_obj;
-    PyObject* send_message_result;
     unsigned char check_keys;
     unsigned char safe;
     unsigned char continue_on_error;
     unsigned char uuid_subtype;
+    unsigned char empty = 1;
     long max_bson_size;
     long max_message_size;
     buffer_t buffer;
@@ -580,6 +618,11 @@ static PyObject* _cbson_do_batched_insert(PyObject* self, PyObject* args) {
     if (continue_on_error) {
         options += 1;
     }
+    /*
+     * If we are doing unacknowledged writes *and* continue_on_error
+     * is True it's pointless (and slower) to send GLE.
+     */
+    send_safe = (safe || !continue_on_error);
 
     max_bson_size_obj = PyObject_GetAttrString(client, "max_bson_size");
 #if PY_MAJOR_VERSION >= 3
@@ -640,31 +683,29 @@ static PyObject* _cbson_do_batched_insert(PyObject* self, PyObject* args) {
         Py_DECREF(doc);
 
         cur_size = buffer_get_position(buffer) - before;
-        max_size = (cur_size > max_size) ? cur_size : max_size;
         if (cur_size > max_bson_size) {
-            PyObject* InvalidDocument = _error("InvalidDocument");
-            if (InvalidDocument) {
-#if PY_MAJOR_VERSION >= 3
-                PyObject* error = PyUnicode_FromFormat(DOC_TOO_LARGE_FMT,
-                                                       cur_size, max_bson_size);
-#else
-                PyObject* error = PyString_FromFormat(DOC_TOO_LARGE_FMT,
-                                                      cur_size, max_bson_size);
-#endif
-                if (error) {
-                    PyErr_SetObject(InvalidDocument, error);
-                    Py_DECREF(error);
-                }
-                Py_DECREF(InvalidDocument);
+            /* If we've encoded anything send it before raising. */
+            if (!empty) {
+                buffer_update_position(buffer, before);
+                message_length = buffer_get_position(buffer) - length_location;
+                memcpy(buffer_get_buffer(buffer) + length_location,
+                       &message_length, 4);
+                result = _send_insert(self, client, last_error_args, buffer,
+                                      collection_name, collection_name_length,
+                                      request_id, send_safe);
+                if (!result)
+                    goto iterfail;
+                Py_DECREF(result);
             }
+            _set_document_too_large(cur_size, max_bson_size);
             goto iterfail;
         }
+        empty = 0;
 
         /* We have enough data, send this batch. */
         if (buffer_get_position(buffer) > max_message_size) {
             int new_request_id = rand();
             int message_start;
-            PyObject* send_gle = Py_False;
             buffer_t new_buffer = buffer_new();
             if (!new_buffer) {
                 PyErr_NoMemory();
@@ -692,29 +733,16 @@ static PyObject* _cbson_do_batched_insert(PyObject* self, PyObject* args) {
             message_length = buffer_get_position(buffer) - length_location;
             memcpy(buffer_get_buffer(buffer) + length_location, &message_length, 4);
 
-            /* If we are doing unacknowledged writes *and* continue_on_error
-             * is True it's pointless (and slower) to send GLE. */
-            if (safe || !continue_on_error) {
-                send_gle = Py_True;
-                if (!add_last_error(self, buffer, request_id, collection_name,
-                                    collection_name_length, last_error_args)) {
-                    buffer_free(new_buffer);
-                    goto iterfail;
-                }
-            }
-            /* Objectify buffer */
-            result = Py_BuildValue("i" BYTES_FORMAT_STRING, request_id,
-                                   buffer_get_buffer(buffer),
-                                   buffer_get_position(buffer));
+            result = _send_insert(self, client, last_error_args, buffer,
+                                  collection_name, collection_name_length,
+                                  request_id, send_safe);
+
             buffer_free(buffer);
             buffer = new_buffer;
             request_id = new_request_id;
             length_location = message_start;
 
-            send_message_result = PyObject_CallMethod(client, "_send_message",
-                                                      "NO", result, send_gle);
-
-            if (!send_message_result) {
+            if (!result) {
                 PyObject *etype = NULL, *evalue = NULL, *etrace = NULL;
                 PyObject* OperationFailure;
                 PyErr_Fetch(&etype, &evalue, &etrace);
@@ -727,8 +755,8 @@ static PyObject* _cbson_do_batched_insert(PyObject* self, PyObject* args) {
                                 /* We're doing unacknowledged writes and
                                  * continue_on_error is False. Just return. */
                                 Py_DECREF(etype);
-                                Py_DECREF(evalue);
-                                Py_DECREF(etrace);
+                                Py_XDECREF(evalue);
+                                Py_XDECREF(etrace);
                                 Py_DECREF(iterator);
                                 buffer_free(buffer);
                                 PyMem_Free(collection_name);
@@ -753,7 +781,7 @@ static PyObject* _cbson_do_batched_insert(PyObject* self, PyObject* args) {
                 PyErr_Restore(etype, evalue, etrace);
                 goto iterfail;
             } else {
-                Py_DECREF(send_message_result);
+                Py_DECREF(result);
             }
         }
     }
@@ -763,7 +791,7 @@ static PyObject* _cbson_do_batched_insert(PyObject* self, PyObject* args) {
         goto insertfail;
     }
 
-    if (!max_size) {
+    if (empty) {
         PyObject* InvalidOperation = _error("InvalidOperation");
         if (InvalidOperation) {
             PyErr_SetString(InvalidOperation, "cannot do an empty bulk insert");
@@ -775,33 +803,21 @@ static PyObject* _cbson_do_batched_insert(PyObject* self, PyObject* args) {
     message_length = buffer_get_position(buffer) - length_location;
     memcpy(buffer_get_buffer(buffer) + length_location, &message_length, 4);
 
-    if (safe) {
-        if (!add_last_error(self, buffer, request_id, collection_name,
-                            collection_name_length, last_error_args)) {
-            goto insertfail;
-        }
-    }
+    /* Send the last (or only) batch */
+    result = _send_insert(self, client, last_error_args, buffer,
+                          collection_name, collection_name_length,
+                          request_id, safe);
 
     PyMem_Free(collection_name);
-
-    /* objectify buffer */
-    result = Py_BuildValue("i" BYTES_FORMAT_STRING, request_id,
-                           buffer_get_buffer(buffer),
-                           buffer_get_position(buffer));
     buffer_free(buffer);
 
-    /* Send the last (or only) batch */
-    send_message_result = PyObject_CallMethod(client, "_send_message", "NN",
-                                              result,
-                                              PyBool_FromLong((long)safe));
-
-    if (!send_message_result) {
+    if (!result) {
         Py_XDECREF(exc_type);
         Py_XDECREF(exc_value);
         Py_XDECREF(exc_trace);
         return NULL;
     } else {
-        Py_DECREF(send_message_result);
+        Py_DECREF(result);
     }
 
     if (exc_type) {
@@ -824,6 +840,358 @@ insertfail:
     return NULL;
 }
 
+static PyObject*
+_send_write_command(PyObject* client, buffer_t buffer,
+                    int lst_len_loc, int cmd_len_loc, unsigned char* errors) {
+
+    PyObject* msg;
+    PyObject* result;
+
+    int request_id = rand();
+    int position = buffer_get_position(buffer);
+    int length = position - lst_len_loc - 1;
+    memcpy(buffer_get_buffer(buffer) + lst_len_loc, &length, 4);
+    length = position - cmd_len_loc;
+    memcpy(buffer_get_buffer(buffer) + cmd_len_loc, &length, 4);
+    memcpy(buffer_get_buffer(buffer), &position, 4);
+    memcpy(buffer_get_buffer(buffer) + 4, &request_id, 4);
+
+    /* objectify buffer */
+    msg = Py_BuildValue("i" BYTES_FORMAT_STRING, request_id,
+                        buffer_get_buffer(buffer),
+                        buffer_get_position(buffer));
+    if (!msg)
+        return NULL;
+
+    /* Send the current batch */
+    result = PyObject_CallMethod(client, "_send_message",
+                                 "NOO", msg, Py_True, Py_True);
+    if (result && PyDict_GetItemString(result, "writeErrors"))
+        *errors = 1;
+    return result;
+}
+
+static buffer_t
+_command_buffer_new(char* ns, int ns_len) {
+    buffer_t buffer;
+    if (!(buffer = buffer_new())) {
+        PyErr_NoMemory();
+        return NULL;
+    }
+    /* Save space for message length and request id */
+    if ((buffer_save_space(buffer, 8)) == -1) {
+        PyErr_NoMemory();
+        buffer_free(buffer);
+        return NULL;
+    }
+    if (!buffer_write_bytes(buffer,
+                            "\x00\x00\x00\x00"  /* responseTo */
+                            "\xd4\x07\x00\x00"  /* opcode */
+                            "\x00\x00\x00\x00", /* options */
+                            12) ||
+        !buffer_write_bytes(buffer,
+                            ns, ns_len + 1) ||  /* namespace */
+        !buffer_write_bytes(buffer,
+                            "\x00\x00\x00\x00"  /* skip */
+                            "\xFF\xFF\xFF\xFF", /* limit (-1) */
+                            8)) {
+        buffer_free(buffer);
+        return NULL;
+    }
+    return buffer;
+}
+
+#define _INSERT 0
+#define _UPDATE 1
+#define _DELETE 2
+
+static PyObject*
+_cbson_do_batched_write_command(PyObject* self, PyObject* args) {
+    struct module_state *state = GETSTATE(self);
+
+    long max_bson_size;
+    long max_cmd_size;
+    long max_write_batch_size;
+    long idx_offset = 0;
+    int idx = 0;
+    int cmd_len_loc;
+    int lst_len_loc;
+    int ns_len;
+    int ordered;
+    char *ns = NULL;
+    PyObject* max_bson_size_obj;
+    PyObject* max_write_batch_size_obj;
+    PyObject* command;
+    PyObject* doc;
+    PyObject* docs;
+    PyObject* client;
+    PyObject* iterator;
+    PyObject* result;
+    PyObject* results;
+    unsigned char op;
+    unsigned char check_keys;
+    unsigned char uuid_subtype;
+    unsigned char empty = 1;
+    unsigned char errors = 0;
+    buffer_t buffer;
+
+    if (!PyArg_ParseTuple(args, "et#bOObbO", "utf-8",
+                          &ns, &ns_len, &op, &command, &docs,
+                          &check_keys, &uuid_subtype, &client)) {
+        return NULL;
+    }
+
+    max_bson_size_obj = PyObject_GetAttrString(client, "max_bson_size");
+#if PY_MAJOR_VERSION >= 3
+    max_bson_size = PyLong_AsLong(max_bson_size_obj);
+#else
+    max_bson_size = PyInt_AsLong(max_bson_size_obj);
+#endif
+    Py_XDECREF(max_bson_size_obj);
+    if (max_bson_size == -1) {
+        PyMem_Free(ns);
+        return NULL;
+    }
+    /*
+     * Max BSON object size + 16k - 2 bytes for ending NUL bytes
+     * XXX: This should come from the server - SERVER-10643
+     */
+    max_cmd_size = max_bson_size + 16382;
+
+    max_write_batch_size_obj = PyObject_GetAttrString(client, "max_write_batch_size");
+#if PY_MAJOR_VERSION >= 3
+    max_write_batch_size = PyLong_AsLong(max_write_batch_size_obj);
+#else
+    max_write_batch_size = PyInt_AsLong(max_write_batch_size_obj);
+#endif
+    Py_XDECREF(max_write_batch_size_obj);
+    if (max_write_batch_size == -1) {
+        PyMem_Free(ns);
+        return NULL;
+    }
+
+    /* Default to True */
+    ordered = !((PyDict_GetItemString(command, "ordered")) == Py_False);
+
+    if (!(results = PyList_New(0))) {
+        PyMem_Free(ns);
+        return NULL;
+    }
+
+    if (!(buffer = _command_buffer_new(ns, ns_len))) {
+        PyMem_Free(ns);
+        Py_DECREF(results);
+        return NULL;
+    }
+
+    PyMem_Free(ns);
+
+    /* Position of command document length */
+    cmd_len_loc = buffer_get_position(buffer);
+    if (!write_dict(state->_cbson, buffer, command, 0, uuid_subtype, 0)) {
+        goto cmdfail;
+    }
+
+    /* Write type byte for array */
+    *(buffer_get_buffer(buffer) + (buffer_get_position(buffer) - 1)) = 0x4;
+
+    switch (op) {
+    case _INSERT:
+        {
+            if (!buffer_write_bytes(buffer, "documents\x00", 10))
+                goto cmdfail;
+            break;
+        }
+    case _UPDATE:
+        {
+            /* MongoDB does key validation for update. */
+            check_keys = 0;
+            if (!buffer_write_bytes(buffer, "updates\x00", 8))
+                goto cmdfail;
+            break;
+        }
+    case _DELETE:
+        {
+            /* Never check keys in a delete command. */
+            check_keys = 0;
+            if (!buffer_write_bytes(buffer, "deletes\x00", 8))
+                goto cmdfail;
+            break;
+        }
+    default:
+        {
+            PyObject* InvalidOperation = _error("InvalidOperation");
+            if (InvalidOperation) {
+                PyErr_SetString(InvalidOperation, "Unknown command");
+                Py_DECREF(InvalidOperation);
+            }
+            goto cmdfail;
+        }
+    }
+
+    /* Save space for list document */
+    lst_len_loc = buffer_save_space(buffer, 4);
+    if (lst_len_loc == -1) {
+        PyErr_NoMemory();
+        goto cmdfail;
+    }
+
+    iterator = PyObject_GetIter(docs);
+    if (iterator == NULL) {
+        PyObject* InvalidOperation = _error("InvalidOperation");
+        if (InvalidOperation) {
+            PyErr_SetString(InvalidOperation, "input is not iterable");
+            Py_DECREF(InvalidOperation);
+        }
+        goto cmdfail;
+    }
+    while ((doc = PyIter_Next(iterator)) != NULL) {
+        int sub_doc_begin = buffer_get_position(buffer);
+        int cur_doc_begin;
+        int cur_size;
+        int enough_data = 0;
+        int enough_documents = 0;
+        char key[16];
+        empty = 0;
+        INT2STRING(key, idx);
+        if (!buffer_write_bytes(buffer, "\x03", 1) ||
+            !buffer_write_bytes(buffer, key, (int)strlen(key) + 1)) {
+            Py_DECREF(doc);
+            goto cmditerfail;
+        }
+        cur_doc_begin = buffer_get_position(buffer);
+        if (!write_dict(state->_cbson, buffer, doc,
+                        check_keys, uuid_subtype, 1)) {
+            Py_DECREF(doc);
+            goto cmditerfail;
+        }
+        Py_DECREF(doc);
+
+        /* We have enough data, maybe send this batch. */
+        enough_data = (buffer_get_position(buffer) > max_cmd_size);
+        enough_documents = (idx >= max_write_batch_size);
+        if (enough_data || enough_documents) {
+            buffer_t new_buffer;
+            cur_size = buffer_get_position(buffer) - cur_doc_begin;
+
+            /* This single document is too large for the command. */
+            if (!idx) {
+                if (op == _INSERT) {
+                    _set_document_too_large(cur_size, max_bson_size);
+                } else {
+                    PyObject* DocumentTooLarge = _error("DocumentTooLarge");
+                    if (DocumentTooLarge) {
+                        /*
+                         * There's nothing intelligent we can say
+                         * about size for update and remove.
+                         */
+                        PyErr_SetString(DocumentTooLarge,
+                                        "command document too large");
+                        Py_DECREF(DocumentTooLarge);
+                    }
+                }
+                goto cmditerfail;
+            }
+
+            if (!(new_buffer = buffer_new())) {
+                PyErr_NoMemory();
+                goto cmditerfail;
+            }
+            /* New buffer including the current overflow document */
+            if (!buffer_write_bytes(new_buffer,
+                (const char*)buffer_get_buffer(buffer), lst_len_loc + 5) ||
+                !buffer_write_bytes(new_buffer, "0\x00", 2) ||
+                !buffer_write_bytes(new_buffer,
+                (const char*)buffer_get_buffer(buffer) + cur_doc_begin, cur_size)) {
+                buffer_free(new_buffer);
+                goto cmditerfail;
+            }
+            /*
+             * Roll the existing buffer back to the beginning
+             * of the last document encoded.
+             */
+            buffer_update_position(buffer, sub_doc_begin);
+
+            if (!buffer_write_bytes(buffer, "\x00\x00", 2))
+                goto cmditerfail;
+
+            result = _send_write_command(client, buffer,
+                                         lst_len_loc, cmd_len_loc, &errors);
+
+            buffer_free(buffer);
+            buffer = new_buffer;
+
+            if (!result)
+                goto cmditerfail;
+
+#if PY_MAJOR_VERSION >= 3
+            result = Py_BuildValue("NN",
+                                   PyLong_FromLong(idx_offset), result);
+#else
+            result = Py_BuildValue("NN",
+                                   PyInt_FromLong(idx_offset), result);
+#endif
+            if (!result)
+                goto cmditerfail;
+
+            PyList_Append(results, result);
+            Py_DECREF(result);
+
+            if (errors && ordered) {
+                Py_DECREF(iterator);
+                buffer_free(buffer);
+                return results;
+            }
+            idx_offset += idx;
+            idx = 0;
+        }
+        idx += 1;
+    }
+    Py_DECREF(iterator);
+
+    if (PyErr_Occurred()) {
+        goto cmdfail;
+    }
+
+    if (empty) {
+        PyObject* InvalidOperation = _error("InvalidOperation");
+        if (InvalidOperation) {
+            PyErr_SetString(InvalidOperation, "cannot do an empty bulk write");
+            Py_DECREF(InvalidOperation);
+        }
+        goto cmdfail;
+    }
+
+    if (!buffer_write_bytes(buffer, "\x00\x00", 2))
+        goto cmdfail;
+
+    result = _send_write_command(client, buffer,
+                                 lst_len_loc, cmd_len_loc, &errors);
+    if (!result)
+        goto cmdfail;
+
+#if PY_MAJOR_VERSION >= 3
+    result = Py_BuildValue("NN", PyLong_FromLong(idx_offset), result);
+#else
+    result = Py_BuildValue("NN", PyInt_FromLong(idx_offset), result);
+#endif
+    if (!result)
+        goto cmdfail;
+
+    buffer_free(buffer);
+
+    PyList_Append(results, result);
+    Py_DECREF(result);
+    return results;
+
+cmditerfail:
+    Py_DECREF(iterator);
+cmdfail:
+    Py_DECREF(results);
+    buffer_free(buffer);
+    return NULL;
+}
+
 static PyMethodDef _CMessageMethods[] = {
     {"_insert_message", _cbson_insert_message, METH_VARARGS,
      _cbson_insert_message_doc},
@@ -835,6 +1203,8 @@ static PyMethodDef _CMessageMethods[] = {
      "create a get more message to be sent to MongoDB"},
     {"_do_batched_insert", _cbson_do_batched_insert, METH_VARARGS,
      "insert a batch of documents, splitting the batch as needed"},
+    {"_do_batched_write_command", _cbson_do_batched_write_command, METH_VARARGS,
+     "execute a batch of insert, update, or delete commands"},
     {NULL, NULL, 0, NULL}
 };
 
