@@ -1,4 +1,4 @@
-# Copyright 2009-2012 10gen, Inc.
+# Copyright 2009-2014 MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -57,22 +57,73 @@ class ConfigurationError(PyMongoError):
 class OperationFailure(PyMongoError):
     """Raised when a database operation fails.
 
+    .. versionadded:: 2.7
+       The :attr:`details` attribute.
+
     .. versionadded:: 1.8
        The :attr:`code` attribute.
     """
 
-    def __init__(self, error, code=None):
-        self.code = code
+    def __init__(self, error, code=None, details=None):
+        self.__code = code
+        self.__details = details
         PyMongoError.__init__(self, error)
 
+    @property
+    def code(self):
+        """The error code returned by the server, if any.
+        """
+        return self.__code
+
+    @property
+    def details(self):
+        """The complete error document returned by the server.
+
+        Depending on the error that occurred, the error document
+        may include useful information beyond just the error
+        message. When connected to a mongos the error document
+        may contain one or more subdocuments if errors occurred
+        on multiple shards.
+        """
+        return self.__details
+
+
+class CursorNotFound(OperationFailure):
+    """Raised while iterating query results if the cursor is
+    invalidated on the server.
+
+    .. versionadded:: 2.7
+    """
+
+
+class ExecutionTimeout(OperationFailure):
+    """Raised when a database operation times out, exceeding the $maxTimeMS
+    set in the query or command option.
+
+    .. note:: Requires server version **>= 2.6.0**
+
+    .. versionadded:: 2.7
+    """
+
 
 class TimeoutError(OperationFailure):
-    """Raised when a database operation times out.
+    """DEPRECATED - will be removed in PyMongo 3.0. See WTimeoutError instead.
 
     .. versionadded:: 1.8
     """
 
 
+class WTimeoutError(TimeoutError):
+    """Raised when a database operation times out (i.e. wtimeout expires)
+    before replication completes.
+
+    With newer versions of MongoDB the `details` attribute may include
+    write concern fields like 'n', 'updatedExisting', or 'writtenTo'.
+
+    .. versionadded:: 2.7
+    """
+
+
 class DuplicateKeyError(OperationFailure):
     """Raised when a safe insert or update fails due to a duplicate key error.
 
@@ -82,6 +133,16 @@ class DuplicateKeyError(OperationFailure):
     """
 
 
+class BulkWriteError(OperationFailure):
+    """Exception class for bulk write errors.
+
+    .. versionadded:: 2.7
+    """
+    def __init__(self, results):
+        OperationFailure.__init__(
+            self, "batch op errors occurred", 65, results)
+
+
 class InvalidOperation(PyMongoError):
     """Raised when a client attempts to perform an invalid operation.
     """
@@ -119,3 +180,8 @@ class ExceededMaxWaiters(Exception):
     """
     pass
 
+
+class DocumentTooLarge(InvalidDocument):
+    """Raised when an encoded document is too large for the connected server.
+    """
+    pass
