@@ -1,4 +1,4 @@
-# Copyright 2011-2012 10gen, Inc.
+# Copyright 2011-2014 MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,7 +14,6 @@
 
 """Test the pymongo common module."""
 
-import os
 import sys
 import unittest
 import warnings
@@ -23,15 +22,23 @@ sys.path[0:0] = [""]
 
 from nose.plugins.skip import SkipTest
 
+from bson.binary import UUIDLegacy, OLD_UUID_SUBTYPE, UUID_SUBTYPE
+from bson.code import Code
 from bson.objectid import ObjectId
 from bson.son import SON
 from pymongo.connection import Connection
 from pymongo.mongo_client import MongoClient
 from pymongo.mongo_replica_set_client import MongoReplicaSetClient
 from pymongo.errors import ConfigurationError, OperationFailure
-from test import host, port, pair
+from test import host, port, pair, version
 from test.utils import drop_collections
 
+have_uuid = True
+try:
+    import uuid
+except ImportError:
+    have_uuid = False
+
 
 class TestCommon(unittest.TestCase):
 
@@ -41,21 +48,21 @@ class TestCommon(unittest.TestCase):
         # manager to test this warning nicely. As we can't do that
         # we must test raising errors before the ignore filter is applied.
         warnings.simplefilter("error", UserWarning)
-        self.assertRaises(UserWarning, lambda:
-                MongoClient(host, port, wtimeout=1000, w=0))
         try:
-            MongoClient(host, port, wtimeout=1000, w=1)
-        except UserWarning:
-            self.fail()
+            self.assertRaises(UserWarning, lambda:
+                    MongoClient(host, port, wtimeout=1000, w=0))
+            try:
+                MongoClient(host, port, wtimeout=1000, w=1)
+            except UserWarning:
+                self.fail()
 
-        try:
-            MongoClient(host, port, wtimeout=1000)
-        except UserWarning:
-            self.fail()
-
-        warnings.resetwarnings()
-
-        warnings.simplefilter("ignore")
+            try:
+                MongoClient(host, port, wtimeout=1000)
+            except UserWarning:
+                self.fail()
+        finally:
+            warnings.resetwarnings()
+            warnings.simplefilter("ignore")
 
         # Connection tests
         c = Connection(pair)
@@ -197,9 +204,9 @@ class TestCommon(unittest.TestCase):
         cursor = coll.find(slave_okay=True)
         self.assertTrue(cursor._Cursor__slave_okay)
 
-        coll.set_lasterror_options(j=True)
-        self.assertEqual({'j': True}, coll.get_lasterror_options())
-        self.assertEqual({'j': True}, coll.write_concern)
+        coll.set_lasterror_options(fsync=True)
+        self.assertEqual({'fsync': True}, coll.get_lasterror_options())
+        self.assertEqual({'fsync': True}, coll.write_concern)
         self.assertEqual({}, db.get_lasterror_options())
         self.assertEqual({}, db.write_concern)
         self.assertFalse(db.safe)
@@ -208,8 +215,8 @@ class TestCommon(unittest.TestCase):
         self.assertFalse(c.safe)
 
         db.set_lasterror_options(w='majority')
-        self.assertEqual({'j': True}, coll.get_lasterror_options())
-        self.assertEqual({'j': True}, coll.write_concern)
+        self.assertEqual({'fsync': True}, coll.get_lasterror_options())
+        self.assertEqual({'fsync': True}, coll.write_concern)
         self.assertEqual({'w': 'majority'}, db.get_lasterror_options())
         self.assertEqual({'w': 'majority'}, db.write_concern)
         self.assertEqual({}, c.get_lasterror_options())
@@ -240,7 +247,148 @@ class TestCommon(unittest.TestCase):
         self.assertTrue(coll.insert({'foo': 'bar'}, fsync=True))
         drop_collections(db)
 
-        warnings.resetwarnings()
+    def test_uuid_subtype(self):
+        if not have_uuid:
+            raise SkipTest("No uuid module")
+
+        self.client = MongoClient(pair)
+        self.db = self.client.pymongo_test
+        coll = self.client.pymongo_test.uuid
+        coll.drop()
+
+        def change_subtype(collection, subtype):
+            collection.uuid_subtype = subtype
+
+        # Test property
+        self.assertEqual(OLD_UUID_SUBTYPE, coll.uuid_subtype)
+        self.assertRaises(ConfigurationError, change_subtype, coll, 7)
+        self.assertRaises(ConfigurationError, change_subtype, coll, 2)
+
+        # Test basic query
+        uu = uuid.uuid4()
+        # Insert as binary subtype 3
+        coll.insert({'uu': uu})
+        self.assertEqual(uu, coll.find_one({'uu': uu})['uu'])
+        coll.uuid_subtype = UUID_SUBTYPE
+        self.assertEqual(UUID_SUBTYPE, coll.uuid_subtype)
+        self.assertEqual(None, coll.find_one({'uu': uu}))
+        self.assertEqual(uu, coll.find_one({'uu': UUIDLegacy(uu)})['uu'])
+
+        # Test Cursor.count
+        self.assertEqual(0, coll.find({'uu': uu}).count())
+        coll.uuid_subtype = OLD_UUID_SUBTYPE
+        self.assertEqual(1, coll.find({'uu': uu}).count())
+
+        # Test remove
+        coll.uuid_subtype = UUID_SUBTYPE
+        coll.remove({'uu': uu})
+        self.assertEqual(1, coll.count())
+        coll.uuid_subtype = OLD_UUID_SUBTYPE
+        coll.remove({'uu': uu})
+        self.assertEqual(0, coll.count())
+
+        # Test save
+        coll.insert({'_id': uu, 'i': 0})
+        self.assertEqual(1, coll.count())
+        self.assertEqual(1, coll.find({'_id': uu}).count())
+        self.assertEqual(0, coll.find_one({'_id': uu})['i'])
+        doc = coll.find_one({'_id': uu})
+        doc['i'] = 1
+        coll.save(doc)
+        self.assertEqual(1, coll.find_one({'_id': uu})['i'])
+
+        # Test update
+        coll.uuid_subtype = UUID_SUBTYPE
+        coll.update({'_id': uu}, {'$set': {'i': 2}})
+        coll.uuid_subtype = OLD_UUID_SUBTYPE
+        self.assertEqual(1, coll.find_one({'_id': uu})['i'])
+        coll.update({'_id': uu}, {'$set': {'i': 2}})
+        self.assertEqual(2, coll.find_one({'_id': uu})['i'])
+
+        # Test Cursor.distinct
+        self.assertEqual([2], coll.find({'_id': uu}).distinct('i'))
+        coll.uuid_subtype = UUID_SUBTYPE
+        self.assertEqual([], coll.find({'_id': uu}).distinct('i'))
+
+        # Test find_and_modify
+        self.assertEqual(None, coll.find_and_modify({'_id': uu},
+                                                     {'$set': {'i': 5}}))
+        coll.uuid_subtype = OLD_UUID_SUBTYPE
+        self.assertEqual(2, coll.find_and_modify({'_id': uu},
+                                                  {'$set': {'i': 5}})['i'])
+        self.assertEqual(5, coll.find_one({'_id': uu})['i'])
+
+        # Test command
+        db = self.client.pymongo_test
+        no_obj_error = "No matching object found"
+        result = db.command('findAndModify', 'uuid',
+                            allowable_errors=[no_obj_error],
+                            uuid_subtype=UUID_SUBTYPE,
+                            query={'_id': uu},
+                            update={'$set': {'i': 6}})
+        self.assertEqual(None, result.get('value'))
+        self.assertEqual(5, db.command('findAndModify', 'uuid',
+                                       update={'$set': {'i': 6}},
+                                       query={'_id': uu})['value']['i'])
+        self.assertEqual(6, db.command('findAndModify', 'uuid',
+                                       update={'$set': {'i': 7}},
+                                       query={'_id': UUIDLegacy(uu)}
+                                      )['value']['i'])
+
+        # Test (inline)_map_reduce
+        coll.drop()
+        coll.insert({"_id": uu, "x": 1, "tags": ["dog", "cat"]})
+        coll.insert({"_id": uuid.uuid4(), "x": 3,
+                     "tags": ["mouse", "cat", "dog"]})
+
+        map = Code("function () {"
+                   "  this.tags.forEach(function(z) {"
+                   "    emit(z, 1);"
+                   "  });"
+                   "}")
+
+        reduce = Code("function (key, values) {"
+                      "  var total = 0;"
+                      "  for (var i = 0; i < values.length; i++) {"
+                      "    total += values[i];"
+                      "  }"
+                      "  return total;"
+                      "}")
+
+        coll.uuid_subtype = UUID_SUBTYPE
+        q = {"_id": uu}
+        if version.at_least(self.db.connection, (1, 7, 4)):
+            result = coll.inline_map_reduce(map, reduce, query=q)
+            self.assertEqual([], result)
+
+        result = coll.map_reduce(map, reduce, "results", query=q)
+        self.assertEqual(0, db.results.count())
+
+        coll.uuid_subtype = OLD_UUID_SUBTYPE
+        q = {"_id": uu}
+        if version.at_least(self.db.connection, (1, 7, 4)):
+            result = coll.inline_map_reduce(map, reduce, query=q)
+            self.assertEqual(2, len(result))
+
+        result = coll.map_reduce(map, reduce, "results", query=q)
+        self.assertEqual(2, db.results.count())
+
+        db.drop_collection("result")
+        coll.drop()
+
+        # Test group
+        coll.insert({"_id": uu, "a": 2})
+        coll.insert({"_id": uuid.uuid4(), "a": 1})
+
+        reduce = "function (obj, prev) { prev.count++; }"
+        coll.uuid_subtype = UUID_SUBTYPE
+        self.assertEqual([],
+                         coll.group([], {"_id": uu},
+                                     {"count": 0}, reduce))
+        coll.uuid_subtype = OLD_UUID_SUBTYPE
+        self.assertEqual([{"count": 1}],
+                         coll.group([], {"_id": uu},
+                                    {"count": 0}, reduce))
 
     def test_write_concern(self):
         c = MongoClient(pair)
